# ===========================================================
# ========================= imports =========================
import numpy as _np
from gnssvod.funcs.constants import _CLIGHT
from gnssvod.funcs.constants import _OMEGA
from gnssvod.geodesy.coordinate import cart2ell, ell2topo
# ===========================================================

def _relativistic_clock(x, y, z, vx, vy, vz):
    """ Calculate relativistic clock correction
    Input: Coordinates [m] (x,y,z) and Velocities (vx,vy,vz) [m/s]
           of satellites in ECEF frame
    Output: Relativistic clock correction in seconds unit
    """
    return -2*(x*vx + y*vy + z*vz)/_CLIGHT**2

def _reception_coord(x_t, y_t, z_t, vx, vy, vz, travel_time):
    """ Calculates satellite coordinates at signal reception time
    Input:
        satellite coordinates at transmission time (x_t, y_t, z_t) [m]
        satellite velocity in ECEF frame [m/s]
        signal travel time calculated from pseudorange [s]
    Output:
        satellite coordinates at signal reception time
    """
    x_r = x_t-vx*travel_time
    y_r = y_t-vy*travel_time
    z_r = z_t-vz*travel_time
    return x_r, y_r, z_r

def _sagnac(x_rec, y_rec, z_rec, x_sat, y_sat, z_sat):
    sagnac_cor = _OMEGA*(x_sat*y_rec-y_sat*x_rec) / _CLIGHT
    return sagnac_cor

def _azel(x_rec,y_rec,z_rec, x_sat, y_sat, z_sat, distance):
    lat_rec, lon_rec, h_rec = cart2ell(x_rec,y_rec,z_rec)
    east, north, up = ell2topo(lat_rec, lon_rec, h_rec) 
    unit_p = _np.matrix([(x_sat-x_rec)/distance, 
                         (y_sat-y_rec)/distance,
                         (z_sat-z_rec)/distance])
    elevation = []
    azimuth = []
    for i in range(len(_np.transpose(unit_p))):
        elevation.append((_np.arcsin(_np.matmul(_np.transpose(unit_p[:,i]),up))))
        azimuth.append(_np.arctan2(_np.matmul(_np.transpose(unit_p[:,i]),east), _np.matmul(_np.transpose(unit_p[:,i]),north)))
        elevation[i] = elevation[i].item()
        azimuth[i]   = azimuth[i].item()
    
    elevation = _np.degrees(elevation)
    azimuth = _np.degrees(azimuth)
    zenith = _np.degrees(_np.pi/2)-elevation
    return (azimuth, elevation, zenith)

def posvel():
    raise Warning("This function will be available in the next release...")


"""
Merge split and stuffs...
"""
# ===========================================================
# ========================= imports =========================
import os
import sys
# ===========================================================

__all__ = ["rinex_merge"]

global _FILEPATH
_FILEPATH = os.path.dirname(os.path.abspath(__file__))

def rinex_merge(station, doy, year, directory = os.getcwd()):
    """
    Parameters:
        station: str
        doy: int
        year: int
        directory: str
    Return:
        None
    Output:
        Merged rinex observation file
    """
    rinexObsFiles=[] 
    file_start = station + str(doy)
    file_end = "."+str(year)[2:]+"o" 
    file_fullName = file_start+"0"+file_end 
    for file in os.listdir(directory):
        if file.startswith(file_start) and file.endswith(file_end):
            rinexObsFiles.append(os.path.join(directory, file))

    if os.path.join(directory,file_fullName) in rinexObsFiles:
        print("The file",file_fullName,"exists in the working directory!")
        while True:
            choice = input("Would you like to overwrite the file [Y/N]? | ")
            if choice.upper() == "Y" or choice.upper() == "Yes":
                rinexObsFiles.remove(os.path.join(directory,file_fullName))
                break
            elif choice.upper() == "N" or choice.upper() == "NO":
                sys.exit("Exiting...")
            else:
                print("Invalid choice | [Y/N]")

    with open(os.path.join(directory, file_fullName), 'w') as rinexMerged:
        with open(rinexObsFiles[0]) as rinexTemp:
            for line in rinexTemp:
                if "TIME OF LAST OBS" not in line:
                    rinexMerged.write(line)

    with open(os.path.join(directory, file_fullName), 'a') as rinexMerged:
        for rinex in range(1, len(rinexObsFiles)):
            with open(rinexObsFiles[rinex]) as rinexTemp:
                header = True
                for line in rinexTemp:
                    if header == True:
                        if "END OF HEADER" not in line:
                            continue
                        else:
                            header = False
                    else:
                        rinexMerged.write(line)



"""
GPS

"""
# ===========================================================
# ========================= imports =========================
import os
import http.client
import urllib.request as url
from pathlib import Path
import datetime
from hatanaka import decompress_on_disk
from dateutil.relativedelta import relativedelta
from gnssvod.funcs.filename import (obsFileName,
                                navFileName, nav3FileName, 
                                obs3FileName)
from gnssvod.funcs.date import (datetime2doy,gpsweekday)
from tqdm import tqdm
# ===========================================================

__all__ = ["get_rinex", "get_rinex3", "get_navigation", "get_clock", "get_sp3", "get_ionosphere"]

server_root = 'ftp://gssc.esa.int/gnss'

def check_internet():
    """ To check if there is an internet connection for FTP downloads """
    connection = http.client.HTTPConnection("www.google.com", timeout=5)
    try:
        connection.request("HEAD", "/")
        connection.close()
        return True
    except:
        connection.close()
        return False

def get_rinex(stationList, date_start, date_finish=None, period='day', Datetime=False, directory=os.getcwd()):
    """
    This function downloads IGS rinex observation file from NASA CDDIS ftp server.
    
    Usage: 
        get_rinex(['mate'],'02-01-2017')
        get_rinex(['mate', 'onsa'],'01-01-2017')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '05-01-2017', period = 'day')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '01-06-2017', period = 'month')
        get_rinex(['mate', 'onsa'], date_start = '01-01-2017', date_finish = '03-01-2017', period = 'month')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '01-01-2018', period = 'year')
    """
    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download RINEX file...')
    
    if Datetime == False:
        date_start = datetime.date(year = int(date_start[-4:]), month = int(date_start[-7:-5]), day = int(date_start[-10:-8]))
        if date_finish != None:
            date_finish = datetime.date(year = int(date_finish[-4:]), month = int(date_finish[-7:-5]), day = int(date_finish[-10:-8]))
    
    timedelta = {'day'   : relativedelta(days   = 1),
                 'month' : relativedelta(months = 1),
                 'year'  : relativedelta(years  = 1)}[period]
    dateList = [date_start] # dates of observation files
    if date_finish != None:
        while dateList[-1] != date_finish:
            date = dateList[-1] + timedelta
            dateList.append(date)

    obsFileDir = 'data/daily' # observation file directory in ftp server
    
    for stationName in stationList:
        for date in dateList:
            doy = datetime2doy(date, string = True)
            fileName = obsFileName(stationName, date, zipped = True)
            # check if the file already exist in the directory
            if os.path.exists(fileName)  == True:
                if os.path.exists(fileName[:-2])  == True:
                    print(fileName[:-2] + " exists in working directory")
                    continue
                else:
                    print(fileName + " exists in working directory | Extracting...")
                    Archive(fileName).extractall(os.getcwd())
                    continue
            file_topath = os.path.join(directory, fileName)
            fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'o', fileName] # file directory
            ftp = '/'.join(fileDir)
            # Download the file
            try:
                print('Downloading:', fileName, end= '')
                url.urlretrieve(ftp, file_topath)
                print(" | Download completed for", fileName, " | Extracting...")
                Archive(fileName).extractall(os.getcwd())
            except:
                raise Warning("Requested file", fileName, "cannot be not found!")

def get_navigation(stationList, date_start, date_finish=None, period='day', Datetime=False, directory=os.getcwd()):
    """
    This function downloads mutli-gnss navigation file (.p) from NASA CDDIS ftp server.
    
    Usage: 
        get_navigation(['mate'],'02-01-2017')
        get_navigation(['mate', 'onsa'],'01-01-2017')
        get_navigation(['mate'], date_start = '01-01-2017', date_finish = '05-01-2017', period = 'day')
        get_navigation(['mate'], date_start = '01-01-2017', date_finish = '01-06-2017', period = 'month')
        get_navigation(['mate', 'onsa'], date_start = '01-01-2017', date_finish = '03-01-2017', period = 'month')
        get_navigation(['mate'], date_start = '01-01-2017', date_finish = '01-01-2018', period = 'year')
    """

    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download rinex file')
    
    if Datetime == False:
        date_start = datetime.date(year = int(date_start[-4:]), month = int(date_start[-7:-5]), day = int(date_start[-10:-8]))
        if date_finish != None:
            date_finish = datetime.date(year = int(date_finish[-4:]), month = int(date_finish[-7:-5]), day = int(date_finish[-10:-8]))
    
    timedelta = {'day'   : relativedelta(days   = 1),
                 'month' : relativedelta(months = 1),
                 'year'  : relativedelta(years  = 1)}[period]
    dateList = [date_start]
    if date_finish != None:
        while dateList[-1] != date_finish:
            date = dateList[-1] + timedelta
            dateList.append(date)

    obsFileDir = 'data/daily'
    
    for stationName in stationList:
        for date in dateList:
            if date >= datetime.date(year=2016,month=1,day=1):
                print("Downloading RINEX3 navigation file...")
                doy = datetime2doy(date, string = True)
                fileName = nav3FileName(stationName, date, zipped = True)
                if os.path.exists(fileName)  == True:
                    if os.path.exists(fileName[:-2])  == True:
                        print(fileName[:-2] + " exists in working directory")
                        continue
                    else:
                        print(fileName + " exists in working directory | Extracting...")
                        Archive(fileName).extractall(os.getcwd())
                        continue
                file_topath = os.path.join(directory, fileName)
                fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'p', fileName]
                ftp = '/'.join(fileDir)
                try:
                    print('Downloading:', fileName, end= '')
                    url.urlretrieve(ftp, file_topath)
                    print(" | Download completed for", fileName, " | Extracting...")
                    Archive(fileName).extractall(os.getcwd())
                except:
                    print("| Requested navigation file", fileName, "cannot be not found! | Checking for IGS Navigation File..." )
                    try:
                        igsFileName = nav3FileName("BRDC", date, zipped = True)
                        
                        if os.path.exists(igsFileName)  == True:
                            if os.path.exists(igsFileName[:-3])  == True:
                                print(igsFileName[:-3] + " exists in working directory")
                                continue
                            else:
                                print(igsFileName + " exists in working directory | Extracting...")
                                Archive(igsFileName).extractall(os.getcwd())
                                continue
                        file_topath = os.path.join(directory, igsFileName)
                        fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'p', igsFileName]
                        ftp = '/'.join(fileDir) 
                        print('Downloading:', igsFileName, end= '')
                        url.urlretrieve(ftp, file_topath)
                        print(" | Download completed for", igsFileName, " | Extracting...")
                        Archive(igsFileName).extractall(os.getcwd())
                    except:
                        raise Warning("IGS Navigation File", igsFileName, "cannot be not found!")
            else:
                print("Downloading RINEX2 navigation file...")
                doy = datetime2doy(date, string = True)
                fileName = navFileName(stationName, date, zipped = True)
                if os.path.exists(fileName)  == True:
                    if os.path.exists(fileName[:-2])  == True:
                        print(fileName[:-2] + " exists in working directory")
                        continue
                    else:
                        print(fileName + " exists in working directory | Extracting...")
                        Archive(fileName).extractall(os.getcwd())
                        continue
                file_topath = os.path.join(directory, fileName)
                fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'n', fileName]
                ftp = '/'.join(fileDir) # FTP link of file
                try:
                    print('Downloading:', fileName, end= '')
                    url.urlretrieve(ftp, file_topath)
                    print(" | Download completed for", fileName, " | Extracting...")
                    Archive(fileName).extractall(os.getcwd())
                except:
                    print("| Requested navigation file", fileName, "cannot be not found! | Checking for IGS Navigation File..." )
                    try:
                        igsFileName = navFileName("brdc", date, zipped = True)
                        if os.path.exists(igsFileName)  == True:
                            if os.path.exists(igsFileName[:-2])  == True:
                                print(fileName[:-2] + " exists in working directory")
                                continue
                            else:
                                print(igsFileName + " exists in working directory | Extracting...")
                                Archive(igsFileName).extractall(os.getcwd())
                                continue
                        file_topath = os.path.join(directory, igsFileName)
                        fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'n', igsFileName]
                        ftp = '/'.join(fileDir)
                        print('Downloading:', igsFileName, end= '')
                        url.urlretrieve(ftp, file_topath)
                        print(" | Download completed for", igsFileName, " | Extracting...")
                        Archive(igsFileName).extractall(os.getcwd())
                    except:
                        raise Warning("IGS Navigation File", igsFileName, "cannot be not found!")


def get_rinex3(stationList, date_start, date_finish=None, period='day', Datetime=False, directory=os.getcwd()):
    """
    This function downloads IGS rinex observation file from NASA CDDIS ftp server.
    
    Usage: 
        get_rinex(['mate'],'02-01-2017')
        get_rinex(['mate', 'onsa'],'01-01-2017')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '05-01-2017', period = 'day')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '01-06-2017', period = 'month')
        get_rinex(['mate', 'onsa'], date_start = '01-01-2017', date_finish = '03-01-2017', period = 'month')
        get_rinex(['mate'], date_start = '01-01-2017', date_finish = '01-01-2018', period = 'year')
    """
    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download RINEX file...')
    
    if Datetime == False:
        date_start = datetime.date(year = int(date_start[-4:]), month = int(date_start[-7:-5]), day = int(date_start[-10:-8]))
        if date_finish != None:
            date_finish = datetime.date(year = int(date_finish[-4:]), month = int(date_finish[-7:-5]), day = int(date_finish[-10:-8]))
    
    timedelta = {'day'   : relativedelta(days   = 1),
                 'month' : relativedelta(months = 1),
                 'year'  : relativedelta(years  = 1)}[period]
    dateList = [date_start]
    if date_finish != None:
        while dateList[-1] != date_finish:
            date = dateList[-1] + timedelta
            dateList.append(date)

    obsFileDir = 'data/daily' # observation file directory in ftp server
    
    for stationName in stationList:
        for date in dateList:
            doy = datetime2doy(date, string = True)
            fileName = obs3FileName(stationName, date, zipped = True)
            # check if the file already exist in the directory
            if os.path.exists(fileName)  == True:
                if os.path.exists(fileName[:-2])  == True:
                    print(fileName[:-2] + " exists in working directory")
                    continue
                else:
                    print(fileName + " exists in working directory | Extracting...")
                    Archive(fileName).extractall(os.getcwd())
                    continue
            file_topath = os.path.join(directory, fileName)
            fileDir = [server_root, obsFileDir, str(date.year), doy, str(date.year)[-2:] + 'd', fileName] 
            ftp = '/'.join(fileDir) 
            try:
                print('Downloading:', fileName, end= '')
                with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=fileName) as t:
                    url.urlretrieve(ftp, file_topath, reporthook=t.update_to)
                print(" | Download completed for", fileName, " | Extracting...")
                Archive(fileName).extractall(os.getcwd())
            except:
                raise Warning("Requested file", fileName, "cannot be not found!")


def get_sp3(sp3_path: str) -> None:
    """
    Given the path of a GFZ orbit file, this function will make sure that orbit file
    is at that location.

    The function will 
    1) do nothing if the file already exists
    2) if a zipped version of the file exists, it will unzip it in the same folder
    3) if no file or zipped file exists, it will download it and unzip it
    """
    sp3path = Path(sp3_path)
    if sp3path.suffix=='.sp3':
        gpsWeek = sp3path.name[3:-5]
        zipped_path = Path(sp3path.parent,sp3path.name + ".Z")
    elif sp3path.suffix=='.SP3':
        sp3Date = datetime.datetime.strptime(sp3path.name.split('_')[1],'%Y%j%H%M').date()
        gpsWeek, gpsWeekday = gpsweekday(sp3Date, Datetime = True)
        gpsWeek = str(gpsWeek)
        zipped_path = Path(sp3path.parent,sp3path.name + ".gz")
    else:
        raise Warning("sp3 filename must either end in .sp3 (gpsWeek < 2238) or .SP3 (gpsWeek >= 2238)")
        sys.exit("Exiting...")
    
    # if the file exists, nothing else to do
    if sp3path.exists():
        print(f"{sp3path.name} exists in {sp3path.parent}")
        return
    # if a zip of the file already exists, unzip it, then leave
    if zipped_path.exists():
        print(f"{zipped_path.name} exists in {zipped_path.parent} | Extracting...")
        decompress_on_disk(zipped_path, delete=True)
        return
    
    # if this is reached, we proceed to downloading
    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download orbit file')
    
    # define remote path
    sp3FileDir = 'products'
    if int(gpsWeek)<2038:
        intermediateDir = ''
    elif int(gpsWeek)<2247: 
        intermediateDir = 'mgex'
    else:
        intermediateDir = ''
    fileDir = [server_root, sp3FileDir, gpsWeek, intermediateDir, zipped_path.name]
    ftp = '/'.join(fileDir) # FTP link of file
    
    # attempt download
    try:
        print('Downloading:', zipped_path.name, end = '')
        with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=zipped_path.name) as t:
            url.urlretrieve(ftp, zipped_path, reporthook=t.update_to)
        print(' | Download completed for', zipped_path.name)
        decompress_on_disk(zipped_path, delete=True)
    except:
        print(" | Requested file", zipped_path.name, "cannot be not found!")

    return
    
def get_clock(clock_path):
    """
    Given the path of a GFZ clock file, this function will make sure that clock file
    is at that location.

    The function will 
    1) do nothing if the file already exists
    2) if a zipped version of the file exists, it will unzip it in the same folder
    3) if no file or zipped file exists, it will download it and unzip it
    """
    clockpath = Path(clock_path)
    if clockpath.suffix=='.clk':
        gpsWeek = clockpath.name[3:-7]
        zipped_path = Path(clockpath.parent,clockpath.name + ".Z")
    elif clockpath.suffix=='.clk_05s':
        gpsWeek = clockpath.name[3:-9]
        zipped_path = Path(clockpath.parent,clockpath.name + ".Z")
    elif clockpath.suffix=='.CLK':
        clockDate = datetime.datetime.strptime(clockpath.name.split('_')[1],'%Y%j%H%M').date()
        gpsWeek, gpsWeekday = gpsweekday(clockDate, Datetime = True)
        gpsWeek = str(gpsWeek)
        zipped_path = Path(clockpath.parent,clockpath.name + ".gz")
    else:
        raise Warning("clock filename must either end in .clk (gpsWeek < 2238) or .CLK (gpsWeek >= 2238)")
        sys.exit("Exiting...")
        
    # if the file exists, nothing else to do
    if clockpath.exists():
        print(f"{clockpath.name} exists in {clockpath.parent}")
        return
    # if a zip of the file already exists, unzip it, then leave
    if zipped_path.exists():
        print(f"{zipped_path.name} exists in {zipped_path.parent} | Extracting...")
        decompress_on_disk(zipped_path, delete=True)
        return
    
    # if this is reached, we proceed to downloading
    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download clock file')
    
    # define remote path
    clockFileDir = 'products'
    if int(gpsWeek)<2038:
        intermediateDir = ''
    elif int(gpsWeek)<2247:
        intermediateDir = 'mgex'
    else:
        intermediateDir = ''
    fileDir = [server_root, clockFileDir, gpsWeek, intermediateDir, zipped_path.name] 
    ftp = '/'.join(fileDir)

    try:
        print('Downloading:', zipped_path.name, end = '')
        with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=zipped_path.name) as t:
            url.urlretrieve(ftp, zipped_path, reporthook=t.update_to)
        print(' | Download completed for', zipped_path.name)
        decompress_on_disk(zipped_path, delete=True)
        return
    except:
        print("Requested file", zipped_path.name, "cannot be not found in ftp server")
        # try alternative name
        altname = "gfz" + zipped_path.name[3:]
        fileDir = [server_root, clockFileDir, altname[3:7], altname] 
        ftp = '/'.join(fileDir)
        try:
            print("Looking for GFZ clock file in ftp server...")
            print('Downloading:', altname, end = '')
            with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=zipped_path.name) as t:
                url.urlretrieve(ftp, zipped_path, reporthook=t.update_to)
            print(' | Download completed for', zipped_path.name)
            decompress_on_disk(zipped_path, delete=True)
        except:
            raise Warning("Requested file", zipped_path.name, "cannot be not found in FTP server | Exiting")

def get_ionosphere(ionFile, directory=os.getcwd()):
    """
    This function downloads Ionosphere file from NASA CDDIS ftp server.
    
    Usage: 
    
    """
    internet = check_internet()
    if internet == False:
        raise Warning('No internet connection! | Cannot download ionosphere file')
    
    fileName = ionFile + ".Z"
    year = int(ionFile[-3:-1])
    if 79 < year < 100:
        year += 1900
    elif year <= 79:
        year += 2000
    # FTP download
    ionFileDir = 'products/ionex'
    file_topath = os.path.join(directory, fileName)
    fileDir = [server_root, ionFileDir, str(year), fileName[4:7], fileName] 
    ftp = '/'.join(fileDir)

    try:
        print('Downloading:', fileName, end = '')
        with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=fileName) as t:
            url.urlretrieve(ftp, file_topath, reporthook=t.update_to)
        print(' | Download completed for', fileName)
        return fileName[:-2]
    except:
        print("Requested file", fileName, "cannot be not found in FTP server")
        fileName = "igs" + ionFile[3:] + ".Z"
        file_topath = os.path.join(directory, fileName)
        fileDir = [server_root, ionFileDir, fileName[3:7], fileName]
        ftp = '/'.join(fileDir)
        try:
            print("Looking for ionosphere file in FTP server...")
            print('Downloading:', fileName, end = '')
            with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1, desc=fileName) as t:
                url.urlretrieve(ftp, file_topath, reporthook=t.update_to)
            print(' | Download completed for', fileName)
            return fileName[:-2]
        except:
            raise Warning("Requested file", fileName, "cannot be not found in FTP server | Exiting")


class TqdmUpTo(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        return self.update(b * bsize - self.n)



# ===========================================================
# ========================= imports =========================
import datetime
# ===========================================================
def gpsweekday(date, Datetime = False):
    start = datetime.date(year= 1980, month= 1, day =6)
    if Datetime == True:
        date = date
    else:
        date = datetime.date(year= int(date[-4:]), month= int(date[-7:-5]), day = int(date[:2]))
    diff = date-start
    diff = diff.days
    week = int(diff/7)
    day = diff%7
    gpswday = int(str(week)+str(day))
    return week, gpswday

def gpswdtodate(gpsweekday):
    week = int(gpsweekday[:4])
    totalday = week*7 + int(gpsweekday[4])
    start = datetime.datetime(year= 1980, month= 1, day =6)
    day = start + datetime.timedelta(days=totalday)
    day = day.date()
    return day

def jday(date):
    """ This function calculates the Julian day of a given date as DD MM YYYY"""
    # year and month
    year = int(date[-4:])
    month = int(date[3:5])
    if month == 1 or month == 2:
        month += 12
        year += -1
    # calculate JD of input time
    A = int(year/100)
    B = int(A/4)
    C = 2-A+B
    D = int(date[:2])
    E = int(365.25*(year+4716))
    F = int(30.6001*(month+1))
    Jd = C+D+E+F-1524.5
    return Jd

def julianday2date(JDay):
    MJd = JDay - 2400000.5 #calculate modified Julian date
    start = datetime.datetime(year= 1858, month = 11, day = 17, hour=0, minute=0, second=0)
    date = start + datetime.timedelta(days=MJd)
    return date

def doy(date: str) -> str:
    """ This function calculates the GPS day of year for the date given """
    year=int(date[-4:])
    month=int(date[-7:-5])
    day=int(date[-10:-8])
    doy=((month-1)*(30))+day
    if year%4!=0:
        if month == 1 or month == 4 or month == 5:
            doy=doy
        elif month == 2 or month == 6 or month == 7:
            doy=doy+1
        elif month == 3:
            doy=doy-1
        elif month == 8:
            doy=doy+2
        elif month == 9 or month == 10:
            doy=doy+3
        elif month ==11 or month == 12:
            doy=doy+4
    #for leap year
    else:
        if month ==1 or month ==3:
            doy=doy
        elif month ==2 or month == 4 or month == 5:
            doy=doy+1
        elif month ==6 or month == 7:
            doy=doy+2
        elif month ==8:
            doy=doy+3
        elif month ==9 or month == 10:
            doy=doy+4
        elif month ==11 or month == 12:
            doy=doy+5
    return doy

def doy2date(rinexFile):
    if len(rinexFile) == 12: # RINEX 2.x naming
        if int(rinexFile[-3:-1]) < 80:
            year = int('20'+rinexFile[-3:-1])
        else:
            year = int('19'+rinexFile[-3:-1])
        doy = int(rinexFile[4:7])
    elif len(rinexFile) == 38: # RINEX 3.x naming
        year = int(rinexFile.split("_")[2][0:4])
        doy  = int(rinexFile.split("_")[2][4:7])
    # Datetime
    start = datetime.date(year = year, month = 1,  day =1)
    date = start + datetime.timedelta(days=doy)-datetime.timedelta(days=1)
    return date

def datetime2doy(date, string = False):
    start = datetime.date(year= date.year, month = 1,  day =1)
    doy = date - start + datetime.timedelta(days = 1)
    doy = doy.days
    if string == True:
        doy = str(doy)
        if len(doy) == 1:
            doy = "00" + doy 
        elif len(doy) == 2:
            doy = "0" + doy
        else:
            doy = doy 
    return doy



# ===========================================================
# ========================= imports =========================
import sys
import datetime
import pathlib
from typing import Union
from gnssvod.funcs.date import (gpsweekday, datetime2doy)
from gnssvod.doc.IGS import IGS, is_IGS
# ===========================================================

def obsFileName(stationName, date, zipped = False):
    doy = datetime2doy(date, string = True)
    if len(doy) == 1:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "o"
    elif len(doy) == 2:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "o"
    else:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "o"
    
    if zipped == True:
        rinexFile = rinexFile + ".Z"
    
    return rinexFile

def sp3FileName(epoch: datetime.date, product: str ="gfz", dir: Union[str,None] = None):
    now = datetime.date.today() # today's date
    timeDif = now - epoch # time difference between rinex epoch and today
    gpsWeek, gpsWeekday = gpsweekday(epoch, Datetime = True)

    if gpsWeek<2038:
        # rely on older gnss orbits
        if len(str(gpsWeek)) == 3:
            sp3File = product.lower() + "0" + str(gpsWeekday) + ".sp3"
        else:
            sp3File = product.lower() + str(gpsWeekday) + ".sp3"
    else:
        if timeDif.days == 0:
            raise Warning("Orbit files are not released yet for", epoch.ctime())
            sys.exit("Exiting...") 
        # we will force using GFZ for now..
        product = 'GFZ0MGXRAP'
        suffix  = '0000_01D_05M_ORB.SP3'
        sp3File = product.upper() + "_" + epoch.strftime('%Y%j') + suffix
    
    if not dir is None:
        sp3File = str(pathlib.Path(dir,sp3File))

    return sp3File

def clockFileName(epoch, interval=30, product="cod", dir: Union[str,None] = None):
    now = datetime.date.today()
    timeDif = now - epoch
    gpsWeek, gpsWeekday = gpsweekday(epoch, Datetime = True) 

    if gpsWeek<2038:
        # rely on older gnss orbits
        if interval < 30:
            product = 'cod'
            extension = '.clk_05s'
        else:
            extension = '.clk'

        if len(str(gpsWeek)) == 3:
            clockFile = product.lower() + "0" + str(gpsWeekday) + extension
        else:
            clockFile = product.lower() + str(gpsWeekday) + extension
    else:
        if timeDif.days == 0:
            raise Warning("Clock files are not released yet for", epoch.ctime())
            sys.exit("Exiting...") 
        # we will force using GFZ for now..
        product = 'GFZ0MGXRAP'
        suffix  = '0000_01D_30S_CLK.CLK'
        clockFile = product.upper() + "_" + epoch.strftime('%Y%j') + suffix

    if not dir is None:
        clockFile = str(pathlib.Path(dir,clockFile))

    return clockFile

def ionFileName(date, product = "igs", zipped = False):
    doy = datetime2doy(date, string = True)
    if len(doy) == 1:
        ionFile = product + "g" + doy + "0." + str(date.year)[-2:] + "i"
    elif len(doy) == 2:
        ionFile = product + "g" + doy + "0." + str(date.year)[-2:] + "i"
    else:
        ionFile = product + "g" + doy + "0." + str(date.year)[-2:] + "i"

    if zipped == True:
        ionFile = ionFile + ".Z"
    
    return ionFile

def navFileName(stationName, date, zipped = False):
    doy = datetime2doy(date, string = True)
    if len(doy) == 1:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "n"
    elif len(doy) == 2:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "n"
    else:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "n"
    
    if zipped == True:
        rinexFile = rinexFile + ".Z"
    
    return rinexFile

def nav3FileName(stationName, date, zipped = False):
    doy = datetime2doy(date, string = True) # for RINEX data names
    siteInfo = IGS(stationName)
    if stationName.upper() == "BRDC":
        rinexFile = "BRDC00IGS_R_" + str(date.year) + str(doy) + "0000_01D_MN.rnx"
    else:
        rinexFile = siteInfo.SITE[0] + "_R_" + str(date.year) + str(doy) + "0000_01D_MN.rnx"
    """
    if len(doy) == 1:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "p"
    elif len(doy) == 2:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "p"
    else:
        rinexFile = stationName + doy + "0." + str(date.year)[-2:] + "p"
    """
    if zipped == True:
        rinexFile = rinexFile + ".gz"
    
    return rinexFile

def obs3FileName(stationName, date, zipped = False):
    doy = datetime2doy(date, string = True) # for RINEX data names
    siteInfo = IGS(stationName)
    rinexFile = siteInfo.SITE[0] + "_R_" + str(date.year) + str(doy) + "0000_01D_30S_MO.crx"
    if zipped == True:
        rinexFile = rinexFile + ".gz"
    
    return rinexFile



"""
Atmopspheric corrections

"""
import numpy as _np
from gnssvod.geodesy.coordinate import cart2ell as _cart2ell
from gnssvod.funcs.date import datetime2doy as _datetime2doy

__all__ = ["tropospheric_delay"]

def tropospheric_delay(x, y, z, elevation, epoch):
    """
    Calculates tropospheric delay using Colins(1999) method
    Input:
        Cartesian coordinates of receiver in ECEF frame (x,y,z)
        Elevation Angle [unit: degree] of satellite vehicle
    Output:
        Tropospheric delay [unit: m]
    
    Reference: 
    Collins, J. P. (1999). Assessment and Development of a Tropospheric Delay Model for
    Aircraft Users of the Global Positioning System. M.Sc.E. thesis, Department of
    Geodesy and Geomatics Engineering Technical Report No. 203, University of
    New Brunswick, Fredericton, New Brunswick, Canada, 174 pp
    """
    lat,lon,ellHeight = _cart2ell(x,y,z)
    ortHeight = ellHeight # Using elliposidal height for now
    # --------------------
    # constants
    k1 = 77.604  # K/mbar
    k2 = 382000  # K^2/mbar
    Rd = 287.054 # J/Kg/K
    g  = 9.80665 # m/s^2
    gm = 9.784   # m/s^2
    # --------------------
    # linear interpolation of meteorological values
    # Average values
    ave_params = _np.array([
                           [1013.25,299.65,26.31,6.30e-3,2.77],
                           [1017.25,294.15,21.79,6.05e-3,3.15],
                           [1015.75,283.15,11.66,5.58e-3,2.57],
                           [1011.75,272.15, 6.78,5.39e-3,1.81],
                           [1013.00,263.65, 4.11,4.53e-3,1.55]
                           ])
    # seasonal variations
    sea_params = _np.array([
                           [ 0.00, 0.00,0.00,0.00e-3,0.00],
                           [-3.75, 7.00,8.85,0.25e-3,0.33],
                           [-2.25,11.00,7.24,0.32e-3,0.46],
                           [-1.75,15.00,5.36,0.81e-3,0.74],
                           [-0.50,14.50,3.39,0.62e-3,0.30]
                           ])
    # Latitude index 
    Latitude = _np.linspace(15,75,5)
    if abs(lat)<=15.0:
        indexLat=0
    elif 15<abs(lat)<=30:
        indexLat=1
    elif 30<abs(lat)<=45:
        indexLat=2
    elif 45<abs(lat)<=60:
        indexLat=3
    elif 60<abs(lat)<75:
        indexLat=4
    elif 75<=abs(lat):
        indexLat=5
    # ----------------
    if indexLat == 0:
        ave_meteo  = ave_params[indexLat,:]
        svar_meteo = sea_params[indexLat-1,:]
    elif indexLat == 5:
        ave_meteo  = ave_params[indexLat-1,:]
        svar_meteo = sea_params[indexLat-1,:]
    else:
        ave_meteo  = ave_params[indexLat-1,:]+(ave_params[indexLat,:]-ave_params[indexLat-1,:])*(abs(lat)-Latitude[indexLat-1])/(Latitude[indexLat]-Latitude[indexLat-1])
        svar_meteo = sea_params[indexLat-1,:]+(sea_params[indexLat,:]-sea_params[indexLat-1,:])*(abs(lat)-Latitude[indexLat-1])/(Latitude[indexLat]-Latitude[indexLat-1])
    #
    doy = _datetime2doy(epoch, string = False)
    if lat >= 0.0: # northern hemisphere
        doy_min=28
    else: # southern latitudes
        doy_min=211
    param_meteo = ave_meteo - svar_meteo*_np.cos((2*_np.pi*(doy-doy_min))/365.25)
    pressure, temperature, e, beta, lamda = param_meteo[0],param_meteo[1],param_meteo[2],param_meteo[3],param_meteo[4]
    # --------------------
    ave_dry = 1e-6*k1*Rd*pressure/gm 
    ave_wet = 1e-6*k2*Rd/(gm*(lamda+1)-beta*Rd)*e/temperature
    d_dry = ave_dry*(1-beta*ortHeight/temperature)**(g/Rd/beta)
    d_wet = ave_wet*(1-beta*ortHeight/temperature)**(((lamda+1)*g/Rd/beta)-1)
    m_elev = 1.001/_np.sqrt(0.002001+_np.sin(_np.deg2rad(elevation))**2)
    dtropo = (d_dry + d_wet) * m_elev
    return dtropo
